@@ -2,7 +2,8 @@
 # -*- coding: utf-8 -*-
 
 from PySide2.QtQml import QQmlApplicationEngine, QQmlContext
-from PySide2.QtCore import QObject, QAbstractListModel, QFileInfo, QFileSystemWatcher, QModelIndex, QDateTime, QDate, QTime, QThreadPool, QThread, QTimer, Qt, QSettings, QStandardPaths, Slot, Signal, Property
+from PySide2.QtCore import QObject, QAbstractListModel, QFileInfo, QFile, QIODevice, QFileSystemWatcher, QModelIndex, QDateTime, QDate, QTime, QThreadPool, QThread, QTimer, Qt, QSettings, QStandardPaths, Slot, Signal, Property
+from PySide2.QtNetwork import QNetworkAccessManager, QNetworkReply, QNetworkRequest
 
 import os.path
 import pathlib
@@ -32,6 +33,15 @@ def human_date(date):
     return QDate.fromString(date, "yyyyMMdd").toString(Qt.RFC2822Date)
 
 
+def singleVideoIfPlaylist(url):
+    index = url.find("&list=")
+
+    if index != -1:
+        url = url[:index]
+
+    return url
+
+
 class PreDownloadTask(QThread):
     collected_info = Signal(dict)
 
@@ -62,6 +72,7 @@ class PreDownload(QObject):
         self.status = "processing"
         self.id = hash(url)
         self.url = url
+        self.destination_file = str()
         self.options = DownloadOptions(options)
 
         self.data = DownloadData()
@@ -83,12 +94,13 @@ class PreDownload(QObject):
             self.task.exit()
             self.task.wait() # NOTE: Should not be costly enough
 
-    def already_downloaded(self):
-        destination_file = "{root}/{title}.{ext}".format(root=self.options.output_path, title=self.data.title, ext=self.options.file_format)
-        return os.path.isfile(destination_file)
+    def destination_file_exists(self):
+        return os.path.isfile(self.destination_file)
 
     def update(self):
-        self.status = "ready" if not self.already_downloaded() else "exists"
+        self.destination_file = os.path.join(self.options.output_path, Paths.new_extension(self.data.title, self.options.file_format))
+
+        self.status = "ready" if not self.destination_file_exists() else "exists"
 
         if self.options.need_post_process():
             self.options.calc_post_process_file_size(self.data._duration)
@@ -110,11 +122,13 @@ class PreDownload(QObject):
 
         if self.task.error:
             self.status = self.task.error
+            self.updated.emit(str(self.id))
 
     @staticmethod
     def pack(predownload):
         return {
             "url": predownload.url,
+            "destination_file": predownload.destination_file,
             "status": predownload.status,
             "data": DownloadData.pack(predownload.data),
             "options": DownloadOptions.pack(predownload.options)
@@ -123,13 +137,14 @@ class PreDownload(QObject):
     @staticmethod
     def unpack(data):
         predownload = PreDownload(data["url"], data["options"])
+        predownload.destination_file = data["destination_file"]
         predownload.status = data["status"]
         predownload.data = DownloadData.unpack(data["data"])
         return predownload
 
 
 class PreDownloadModel(QAbstractListModel):
-    COLUMNS = ("url", "status", "download_data", "options")
+    COLUMNS = ("url", "destination_file", "status", "download_data", "options")
     FIRST_COLUMN = 0
     LAST_COLUMN = len(COLUMNS)
 
@@ -179,9 +194,10 @@ class PreDownloadModel(QAbstractListModel):
     def roleNames(self, index=QModelIndex()):
         return {
             256: b"url",
-            257: b"status",
-            258: b"download_data",
-            259: b"options"
+            257: b"destination_file",
+            258: b"status",
+            259: b"download_data",
+            260: b"options"
         }
 
     def index(self, row, column, parent):
@@ -238,12 +254,15 @@ class PreDownloadModel(QAbstractListModel):
             return predownload.url
 
         elif role == 257:
-            return predownload.status
+            return predownload.destination_file
 
         elif role == 258:
-            return predownload.data
+            return predownload.status
 
         elif role == 259:
+            return predownload.data
+
+        elif role == 260:
             return predownload.options
 
         return None
@@ -257,7 +276,7 @@ class PreDownloadModel(QAbstractListModel):
 
         # NOTE: Do we want to add option for url changing?
 
-        if role == 259:
+        if role == 260:
             predownload.options.update(value.toVariant())
             predownload.update()
             self.dataChanged.emit(self.index(row, PreDownloadModel.FIRST_COLUMN, QModelIndex()), self.index(row, PreDownloadModel.LAST_COLUMN, QModelIndex()))
@@ -375,6 +394,10 @@ class DownloadProgress(QObject):
     def downloadSpeed(self):
         return self.speed
 
+    def invalide(self):
+        self.status = "no file"
+        self.downloaded_bytes = 0
+
     @Slot(dict)
     def update(self, data):
         if "status" in data:
@@ -588,12 +611,19 @@ class Download(QObject):
         self.progress = DownloadProgress()
         self.task = DownloadTask(self.url, self.options)
 
+        self.destination_file = os.path.join(self.options.output_path, Paths.new_extension(self.data.title, self.options.file_format))
+
         self.task.progress.connect(self.update, Qt.QueuedConnection)
         self.task.finished.connect(self.handle_finished)
 
+
     def __eq__(self, other):
         return self.url == other.url and self.options == other.options
 
+    def check_if_redownload_needed(self):
+        if not os.path.isfile(self.destination_file) and self.progress.status == "finished":
+            self.progress.invalide()
+
     def start(self):
         if self.running():
             return
@@ -632,6 +662,7 @@ class Download(QObject):
     def pack(download):
         return {
             "url": download.url,
+            "destination_file": download.destination_file,
             "options": DownloadOptions.pack(download.options),
             "data": DownloadData.pack(download.data),
             "progress": DownloadProgress.pack(download.progress),
@@ -641,6 +672,7 @@ class Download(QObject):
     def unpack(data):
         download =  Download(data["url"], data["options"], data["data"])
         download.progress = DownloadProgress.unpack(data["progress"])
+        download.check_if_redownload_needed()
         return download
 
     @classmethod
@@ -673,7 +705,7 @@ class DownloadPostProcess(QObject):
 
 
 class DownloadModel(QAbstractListModel):
-    COLUMNS = ("url", "download_data", "progress", "options")
+    COLUMNS = ("url", "destination_file", "download_data", "progress", "options")
     FIRST_COLUMN = 0
     LAST_COLUMN = len(COLUMNS)
 
@@ -741,9 +773,10 @@ class DownloadModel(QAbstractListModel):
     def roleNames(self, index=QModelIndex()):
         return {
             256: b"url",
-            257: b"download_data",
-            258: b"progress",
-            259: b"options"
+            257: b"destination_file",
+            258: b"download_data",
+            259: b"progress",
+            260: b"options"
         }
 
     def clear(self):
@@ -799,12 +832,15 @@ class DownloadModel(QAbstractListModel):
             return download.url
 
         elif role == 257:
-            return download.data
+            return download.destination_file
 
         elif role == 258:
-            return download.progress
+            return download.data
 
         elif role == 259:
+            return download.progress
+
+        elif role == 260:
             return download.options
 
         return None
@@ -831,7 +867,7 @@ class DownloadManager(QObject):
     @Slot(str, "QVariantMap")
     def predownload(self, url, options):
         if url:
-            predownload = PreDownload(url, options)
+            predownload = PreDownload(singleVideoIfPlaylist(url), options)
             self.predownload_model.add_predownload(predownload)
             predownload.start()
 
@@ -853,3 +889,88 @@ class DownloadManager(QObject):
     def setQMLContext(self, engine):
         engine.rootContext().setContextProperty("downloadModel", self.download_model)
         engine.rootContext().setContextProperty("predownloadModel", self.predownload_model)
+
+
+class FileDownload(QObject):
+    def __init__(self, manager: QNetworkAccessManager, url: str, output_url: str):
+        super(FileDownload, self).__init__(None)
+
+        request = QNetworkRequest(url)
+
+        self.output_url =  output_url
+        self.output = QFile(self.output_url)
+
+        self.logger = create_logger(__name__)
+        if not self.output.open(QIODevice.WriteOnly):
+            self.logger.info("Could not open: {url}".format(url=self.output_url))
+            return
+
+        self.current_download = manager.get(request)
+        self.current_download_progress = FileDownloadProgress()
+        self.current_download.downloadProgress.connect(self.current_download_progress.update)
+        self.current_download.readyRead.connect(self.saveFile)
+        self.current_download.finished.connect(self.download_finished)
+
+    @Slot()
+    def saveFile(self):
+        self.output.write(self.current_download.readAll())
+
+    @Slot()
+    def download_finished(self):
+        self.output.close()
+
+    @Property(str, constant=True)
+    def outputUrl(self):
+        return self.output_url
+
+    @Property(QObject, constant=True)
+    def progress(self):
+        return self.current_download_progress
+
+
+class FileDownloadProgress(QObject):
+    updated = Signal()
+
+    def __init__(self):
+        super(FileDownloadProgress, self).__init__(None)
+
+        self.read_bytes = int()
+        self.total_bytes = int()
+
+    @Slot(int, int)
+    def update(self, read_bytes: int, total_bytes: int):
+        self.read_bytes = read_bytes
+        self.total_bytes = total_bytes
+        self.updated.emit()
+
+    @Property(int, notify=updated)
+    def readBytes(self):
+        return self.read_bytes
+
+    @Property(int, notify=updated)
+    def totalBytes(self):
+        return self.total_bytes
+
+
+class FileDownloader(QObject):
+    currentDownloadChanged = Signal()
+
+    def __init__(self):
+        super(FileDownloader, self).__init__(None)
+
+        self.manager = QNetworkAccessManager()
+        self.current_download = None
+
+    @Slot()
+    def clear(self):
+        self.current_download = None
+        self.currentDownloadChanged.emit()
+
+    @Slot(str, str)
+    def download(self, url: str, output_url: str):
+        self.current_download = FileDownload(self.manager, url, output_url)
+        self.currentDownloadChanged.emit()
+
+    @Property("QVariant", notify=currentDownloadChanged)
+    def currentDownload(self):
+        return self.current_download
